@@ -47,10 +47,7 @@ static inline unsigned long size_inside_page(unsigned long start,
 #ifndef ARCH_HAS_VALID_PHYS_ADDR_RANGE
 static inline int valid_phys_addr_range(unsigned long addr, size_t count)
 {
-	if (addr + count > __pa(high_memory))
-		return 0;
-
-	return 1;
+	return addr + count <= __pa(high_memory);
 }
 
 static inline int valid_mmap_phys_addr_range(unsigned long pfn, size_t size)
@@ -788,10 +785,11 @@ static const struct file_operations zero_fops = {
 /*
  * capabilities for /dev/zero
  * - permits private mappings, "copies" are taken of the source of zeros
+ * - no writeback happens
  */
 static struct backing_dev_info zero_bdi = {
 	.name		= "char/mem",
-	.capabilities	= BDI_CAP_MAP_COPY,
+	.capabilities	= BDI_CAP_MAP_COPY | BDI_CAP_NO_ACCT_AND_WRITEBACK,
 };
 
 static const struct file_operations full_fops = {
@@ -804,62 +802,10 @@ static const struct file_operations full_fops = {
 static const struct file_operations oldmem_fops = {
 	.read	= read_oldmem,
 	.open	= open_oldmem,
+	.llseek = default_llseek,
 };
 #endif
 
-#ifdef CONFIG_SONOS_FILLMORE
-static loff_t memory_arlseek(struct file * file, loff_t offset, int orig)
-{
-        loff_t ret;
-
-        mutex_lock(&file->f_path.dentry->d_inode->i_mutex);
-        switch (orig) {
-                case 0:
-                        file->f_pos = offset & 0xffffffff;
-                        ret = file->f_pos;
-                        force_successful_syscall_return();
-                        break;
-                case 1:
-                        file->f_pos += offset;
-                        file->f_pos &= 0xffffffff;
-                        ret = file->f_pos;
-                        force_successful_syscall_return();
-                        break;
-                default:
-                        ret = -EINVAL;
-        }
-        mutex_unlock(&file->f_path.dentry->d_inode->i_mutex);
-        return ret;
-}
-
-static ssize_t read_armem(struct file *file, char __user *buf, 
-			 size_t count, loff_t *ppos)
-{
-	unsigned int	pos = (unsigned int)*ppos;
-	if (copy_to_user(buf, (void *)pos, count))
-		return -EFAULT;
- 	*ppos += count;
- 	return count;
-}
-
-static ssize_t write_armem(struct file * file, const char __user * buf, 
-			  size_t count, loff_t *ppos)
-{
-	unsigned int	pos = (unsigned int)*ppos;
-	if(copy_from_user((void *)pos, buf, count)) {
-		return -EFAULT;
-	}
-	*ppos += count;
- 	return count;
-}
-
-static struct file_operations armem_fops = {
-	.llseek		= memory_arlseek,
-	.read		= read_armem,
-	.write		= write_armem,
-};
-#endif	// CONFIG_SONOS_FILLMORE
-
 static ssize_t kmsg_write(struct file *file, const char __user *buf,
 			  size_t count, loff_t *ppos)
 {
@@ -883,6 +829,7 @@ static ssize_t kmsg_write(struct file *file, const char __user *buf,
 
 static const struct file_operations kmsg_fops = {
 	.write = kmsg_write,
+	.llseek = noop_llseek,
 };
 
 static const struct memdev {
@@ -891,7 +838,9 @@ static const struct memdev {
 	const struct file_operations *fops;
 	struct backing_dev_info *dev_info;
 } devlist[] = {
+#ifndef SONOS_NO_DEVMEM
 	 [1] = { "mem", 0, &mem_fops, &directly_mappable_cdev_bdi },
+#endif
 #ifdef CONFIG_DEVKMEM
 	 [2] = { "kmem", 0, &kmem_fops, &directly_mappable_cdev_bdi },
 #endif
@@ -907,9 +856,6 @@ static const struct memdev {
 #ifdef CONFIG_CRASH_DUMP
 	[12] = { "oldmem", 0, &oldmem_fops, NULL },
 #endif
-#ifdef CONFIG_SONOS_FILLMORE
-	[13] = { "armem",  S_IRUSR | S_IWUSR | S_IRGRP, &armem_fops, NULL },
-#endif
 };
 
 static int memory_open(struct inode *inode, struct file *filp)
@@ -929,6 +875,10 @@ static int memory_open(struct inode *inode, struct file *filp)
 	if (dev->dev_info)
 		filp->f_mapping->backing_dev_info = dev->dev_info;
 
+	/* Is /dev/mem or /dev/kmem ? */
+	if (dev->dev_info == &directly_mappable_cdev_bdi)
+		filp->f_mode |= FMODE_UNSIGNED_OFFSET;
+
 	if (dev->fops->open)
 		return dev->fops->open(inode, filp);
 
@@ -937,6 +887,7 @@ static int memory_open(struct inode *inode, struct file *filp)
 
 static const struct file_operations memory_fops = {
 	.open = memory_open,
+	.llseek = noop_llseek,
 };
 
 static char *mem_devnode(struct device *dev, mode_t *mode)
@@ -969,10 +920,10 @@ static int __init chr_dev_init(void)
 		if (!devlist[minor].name)
 			continue;
 		device_create(mem_class, NULL, MKDEV(MEM_MAJOR, minor),
-			      NULL, devlist[minor].name);
+			      NULL, "%s", devlist[minor].name);
 	}
 
-	return 0;
+	return tty_init();
 }
 
 fs_initcall(chr_dev_init);
