@@ -282,6 +282,7 @@ enum rtattr_type_t {
 	RTA_SESSION, /* no longer used */
 	RTA_MP_ALGO, /* no longer used */
 	RTA_TABLE,
+	RTA_MARK,
 	__RTA_MAX
 };
 
@@ -492,48 +493,48 @@ enum {
 #define TCA_PAYLOAD(n) NLMSG_PAYLOAD(n,sizeof(struct tcmsg))
 
 /**************************************************************
- *     Rincon Wireless Messages and Attributes
+ *       Rincon Wireless Messages and Attributes
  ****/
 
 enum
 {
-    RWM_ASSOC,
-    RWM_PROBE,
-    RWM_HHID,
-    RWM_MII,
-    RWM_INFRA_NEIGHBOR,
-    RWM_NETSTART_STA,
-    RWM_CLIENT_ASSOC,
-    RWM_SUPPLICANT,
-    RWM_SONOSNET_PATH,
-    RWM_RENEW_DHCP,
-    RWM_SONOSNET_FULL_MESH,
-    RWM_PRIMARY_FOUND,
-    RWM_PRIMARY_LOST,
-    RWM_DUPE_IP,
-    RWM_HTAP_CONN,
-    RWM_CMD_TIMEOUT,
-    RWM_NETWORK_SETTINGS,
+	RWM_ASSOC,
+	RWM_PROBE,
+	RWM_HHID,
+	RWM_MII,
+	RWM_INFRA_NEIGHBOR,
+	RWM_NETSTART_STA,
+	RWM_CLIENT_ASSOC,
+	RWM_SUPPLICANT,
+	RWM_SONOSNET_PATH,
+	RWM_RENEW_DHCP,
+	RWM_SONOSNET_FULL_MESH,
+	RWM_PRIMARY_FOUND,
+	RWM_PRIMARY_LOST,
+	RWM_DUPE_IP,
+	RWM_HTAP_CONN,
+	RWM_CMD_TIMEOUT,
+	RWM_NETWORK_SETTINGS,
 };
 
 enum
 {
-    RWA_DEV_NAME,
-    RWA_ASSOC_MAC,
-    RWA_ASSOC_UUID,
-    RWA_PROBE_MAC,
-    RWA_PROBE_DATA,
-    RWA_HHID,
-    RWA_INFRA_NEIGHBOR_MAC,
-    RWA_RSSI,
-    RWA_RSSI_PER_CHAIN,
-    RWA_NETSTART_STA_MAC,
-    RWA_CLIENT_ASSOC_DATA,
-    RWA_SUPPLICANT_EVENT,
-    RWA_SUPPLICANT_EVENT_DATA,
-    RWA_PATH_LENGTH,
-    RWA_DUPE_TIME,
-    RWA_NETWORK_SETTINGS_DATA,
+	RWA_DEV_NAME,
+	RWA_ASSOC_MAC,
+	RWA_ASSOC_UUID,
+	RWA_PROBE_MAC,
+	RWA_PROBE_DATA,
+	RWA_HHID,
+	RWA_INFRA_NEIGHBOR_MAC,
+	RWA_RSSI,
+	RWA_RSSI_PER_CHAIN,
+	RWA_NETSTART_STA_MAC,
+	RWA_CLIENT_ASSOC_DATA,
+	RWA_SUPPLICANT_EVENT,
+	RWA_SUPPLICANT_EVENT_DATA,
+	RWA_PATH_LENGTH,
+	RWA_DUPE_TIME,
+	RWA_NETWORK_SETTINGS_DATA,
 };
 
 /********************************************************************
@@ -633,7 +634,6 @@ enum rtnetlink_groups {
 	__RTNLGRP_MAX
 };
 #define RTNLGRP_MAX	(__RTNLGRP_MAX - 1)
-
 #ifdef __KERNEL__
 #define RTMGRP_Rincon   RTNLGRP_RINCON
 #else
@@ -656,6 +656,7 @@ struct tcamsg {
 #ifdef __KERNEL__
 
 #include <linux/mutex.h>
+#include <linux/netdevice.h>
 
 static __inline__ int rtattr_strcmp(const struct rtattr *rta, const char *str)
 {
@@ -800,6 +801,35 @@ extern int rtnl_is_locked(void);
 extern int lockdep_rtnl_is_held(void);
 #endif /* #ifdef CONFIG_PROVE_LOCKING */
 
+/**
+ * rcu_dereference_rtnl - rcu_dereference with debug checking
+ * @p: The pointer to read, prior to dereferencing
+ *
+ * Do an rcu_dereference(p), but check caller either holds rcu_read_lock()
+ * or RTNL. Note : Please prefer rtnl_dereference() or rcu_dereference()
+ */
+#define rcu_dereference_rtnl(p)					\
+	rcu_dereference_check(p, rcu_read_lock_held() ||	\
+				 lockdep_rtnl_is_held())
+
+/**
+ * rtnl_dereference - fetch RCU pointer when updates are prevented by RTNL
+ * @p: The pointer to read, prior to dereferencing
+ *
+ * Return the value of the specified RCU-protected pointer, but omit
+ * both the smp_read_barrier_depends() and the ACCESS_ONCE(), because
+ * caller holds RTNL.
+ */
+#define rtnl_dereference(p)					\
+	rcu_dereference_protected(p, lockdep_rtnl_is_held())
+
+static inline struct netdev_queue *dev_ingress_queue(struct net_device *dev)
+{
+	return rtnl_dereference(dev->ingress_queue);
+}
+
+extern struct netdev_queue *dev_ingress_queue_create(struct net_device *dev);
+
 extern void rtnetlink_init(void);
 extern void __rtnl_unlock(void);
 
