@@ -64,7 +64,7 @@ void br_mcast_put(struct net_bridge_mcast_entry *me)
 }
 
 struct net_bridge_mcast_entry *br_mcast_get(struct net_bridge *br, 
-                                            unsigned char *addr)
+                                            const unsigned char *addr)
 {
     struct net_bridge_mcast_entry *me = 0;
 
