@@ -198,6 +198,8 @@ void br_uplink_xmit(struct net_bridge *br,
 {
 	struct net_bridge_port *p_uplink, *p_dst;
 	struct sk_buff *skb2;
+	struct net_bridge_mcast_entry *me;
+	struct net_bridge_mcast_rx_port *mrxp;
 
 #ifdef BR_DEBUG_UPLINK
 	printk("brup: tx: p=%04x d=%02x:%02x:%02x:%02x:%02x:%02x\n",
@@ -211,7 +213,7 @@ void br_uplink_xmit(struct net_bridge *br,
         }
 #endif
 
-	/* 
+	/*
 	 * Unicast
 	 */
 	if (0 == (dest[0] & 1)) {
@@ -226,7 +228,7 @@ void br_uplink_xmit(struct net_bridge *br,
 		return;
 	}
 
-	/* 
+	/*
 	 * Broadcast
 	 *
 	 * This is a little harder.  If it is a known multicast group we
@@ -236,8 +238,8 @@ void br_uplink_xmit(struct net_bridge *br,
 	 *         should filter here as well?
 	 */
 	p_uplink = _get_uplink(br);
-        
-	struct net_bridge_mcast_entry *me = br_mcast_get(br, dest);
+
+	me = br_mcast_get(br, dest);
 
 	if (NULL == me) {
 
@@ -285,7 +287,7 @@ void br_uplink_xmit(struct net_bridge *br,
 
 		} else {
 			br->statistics.tx_dropped++;
-			kfree_skb(skb);	        
+			kfree_skb(skb);
 		}
 		return;
 	}
@@ -295,10 +297,11 @@ void br_uplink_xmit(struct net_bridge *br,
 #endif
 
 	/* Walk the multicast entry, unicasting to each destination */
-	struct net_bridge_mcast_rx_port* mrxp;
 	for (mrxp = me->rx_port_list; mrxp != NULL; mrxp = mrxp->next) {
 
 		struct net_bridge_port* p = mrxp->dst;
+		struct net_bridge_mcast_rx_mac *mrxm;
+		struct net_bridge_port *p_dst;
 
 		/* Skip local ports */
 		if (!p) {
@@ -308,9 +311,6 @@ void br_uplink_xmit(struct net_bridge *br,
 		/* Walk the MACs for each port (should all be in the
 		 * uplink port, so there should only be one).
 		 */
-		struct net_bridge_mcast_rx_mac* mrxm;
-		struct net_bridge_port* p_dst;
-			
 		for (mrxm = mrxp->rx_mac_list;
 		     mrxm != NULL;
 		     mrxm = mrxm->next) {
@@ -536,6 +536,7 @@ struct sk_buff *br_uplink_handle_frame(struct net_bridge_port *p,
 				       const unsigned char *dst)
 {
 	struct net_bridge *br = p->br;
+	int passup = 0;
 
 	/*
 	 * REVIEW: Optimize by dropping multicasts we don't care about?
@@ -545,7 +546,7 @@ struct sk_buff *br_uplink_handle_frame(struct net_bridge_port *p,
 	 * are not part of would be dropped, but what do I know?
 	 *
 	 */
-	
+
 	/* Always check for multicast table updates */
 	br_mcast_check(skb, p->br, p);
 
@@ -567,15 +568,13 @@ struct sk_buff *br_uplink_handle_frame(struct net_bridge_port *p,
 	       dst[0], dst[1], dst[2],
 	       dst[3], dst[4], dst[5],
 	       p->port_no,
-	       skb->len);                        
+	       skb->len);
 #endif
 
 	/*
 	 * Unless proxying, we either pass up or drop. Only proxied
 	 * frames are forwarded toward satellites.
 	 */
-	int passup = 0;
-
 	if (dst[0] & 1 || _br_uplink_handle_ipv4_broadcast(br, skb)) {
 		passup = 1;
 #ifdef CONFIG_SONOS_BRIDGE_PROXY
